/* SPDX-License-Identifier: GPL-2.0-only
 * Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License v2. See the file COPYING in the main directory of this archive for
 * more details.
 */

#ifndef EVDI_PARAMS_H
#define EVDI_PARAMS_H

extern unsigned int evdi_loglevel;
extern unsigned short int evdi_initial_device_count;

#endif /* EVDI_PARAMS_H */
